/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

//
// File purpose: Example of using Array<N,T> objects
//

#include <iostream>
#include "expresso/expresso.h"

using namespace expresso;
using std::cout;
using std::endl;

int main() {
  // The array class name is Array<N,Type> where N is rank and Type
  // identifies the implementation and the element type.
  // In this file we will use the implementation Array<N,T>
  // where T is the element type.

  // Note that almost no runtime checks of the parameters exist.
  // If you want runtime checks, compile with -DEXPRESSO_DEBUG_ALL
  // The code will then be slower, but almost all parameters are checked
  // at runtime and exceptions are thrown for errors

  ///
  /// Constructors
  ///

  // The following constructors exist

  Array<2,float> a1;     
  // Creates an array object with rank 2 and element type float which 
  // is initially empty, meaning that no memory for elements is 
  // allocated initially.

  Array<2,float> a2(15,25);
  // Creates an array object with rank 2 and element type float with
  // index range [0:14,0:24]

  Ix<2> len(10,20);
  Array<2,float> a3(len);
  // Creates an array object with rank 2 and element type int with
  // index range [0:9,0:19]

  Ix<2> bottom(1,5);
  Array<2,float> a4(len,bottom);
  // Creates an array object with rank 2 and element type int with
  // index range [1:10,5:24]

  Array<2,float> a5(a4);
  // Copy constructor. The array a5 is initialized with the same
  // size as a4 and pointing to the same element data memory. This
  // means that modifying elements of a5 will modify the elements
  // of a4 and vice versa. 

  Array<2,float> a6 = a4;
  // Equivalent to Array<2,float> a6(a4);

  // There also exists template constructors which initializes
  // an array from an array of another type, more about that later

  Array<2,const float> a7 = a4;
  // This creates an array whose elements may not be modified.
  // It is initialized to point to the same memory area as a4.

  // The following is not allowed, because it would allow the elements
  // of the  Array<N,const T> object a7 to be modified by 
  // first assigning it to a Array<N,T> object

  // Array<2,float> a8 = a7;   // Error!!!

  ///
  /// Member access functions
  ///

  for(int n=0; n<2; n++)
    cout << a4.L(n) <<' '<<a4.U(n)<<' '<<a4.Len(n)<<' '<<a4.Stride(n)<<endl;
  // Should produce the output
  // 1 11 10 1
  // 5 25 20 10
  // These member functions have the following meaning
  // L(n)         The first index inside allowed range for rank n
  // U(n)         The first index outside allowed range for rank n
  // Len(n)       Short for U(n)-L(n)
  // Stride(n)    The stride (number of elements in memory) between
  //              two elements along rank n. This can be negative or zero.

  cout<<(void *)a4.Start()<<endl;
  // Returns the address of the first element

  cout<<(void *)a4.Origo()<<endl;
  // Returns the address of the element at index [0,0]. In this
  // case the index range of a4 is [1:10,5:24] so [0,0] is outside
  // the index range. Nevertheless, a4.Origo() returns an address
  // which can be used as base.

  a4.Data();
  // This function returns a low level object which keeps track of
  // allocated element memory area. Normally you should not need to
  // call this member function.

  cout<<a4.IsNull()<<endl;
  // returns true iff a4 is empty

  ///
  /// Indexing member operator
  ///
  
  a4(6,8)=3.14;
  cout<<a4(6,8)<<endl;
  // The ()-operators are used for indexing. a4(6,8) returns a 
  // reference to the element at index position [6,8]. If the element
  // type is not a const type this reference can be assigned to

  Ix<2> index(6,8);
  cout<<a4[index]<<endl;
  // Ix<N> objects may also be used to pick out an element from an
  // array

  // Note that the indexing operators do not check that the index
  // is inside the valid index range of the array. The result of
  // accessing an element outside this range is undefined. However,
  // if you compile the code with -DEXPRESSO_DEBUG_ALL or alternatively
  // with -DEXPRESSO_DEBUG_CHECK_INDEX_RANGE the index
  // range is checked, and an exception is thrown if the index
  // operators are called with a nonvalid index

  ///
  /// Manipulation member functions
  ///

  // To reinitialize an array with a new memory area and index range,
  // use the member function Reinit. Several variations exist, notice
  // that the parameter format is the same as in corresponding constructors
  // except in the last case:

  a4.Reinit();
  // After this the array is empty (no memory allocated for elements)

  a4.Reinit(len);
  // The a new memory block is allocated and the index range is 
  // [0:len[0]-1,0:len[1]-1]

  a4.Reinit(len, bottom);
  // As above, but new index range is
  // [bottom[0]:bottom[0]+len[0]-1, bottom[1]:bottom[1]+len[1]-1]

  // For convenience the following version of Reinit exists:
  Array<1,float> a8 = Ix<1>(len[0]*len[1]);
  a4.Reinit(len, bottom, a8.Data(), a8.Origo()-bottom[0]-bottom[1]*len[0]);
  // In this example, a rank-1 array a8 is created whose element memory area
  // has length len[0]*len[1]. Then a4 is reinited to point to the same
  // element memory area but organized as a rank-2 array with low
  // index bound bottom. Note that the origo of a8 must be modified
  // because the low index bottom is not zero.
  // If you decide to use this function you must make sure that the
  // element memory are is large enough for the index range and that
  // the origo value is correct, there is no way to determine this
  // by runtime checks inside expresso.
  // Use only when necessary.

  a3.Reinit(a4);
  // Here a3 is reinited with a memory block of the same size 
  // as a4.
  // This is not equivalent to the constructor a3(a4) where a3 and a4
  // will point to the same element memory. If you want that effect use the
  // following:
  a3.Refer(a4);

  a3.Clone(a4);
  // Clone works as Reinit(a4), but the elements of a4 are also copied
  // into the newly allocated memory of a3.

  a3.Clone();
  // This is equivalent to a3.Clone(a3)

  // The following "logical operators" exist for manipulating index range:

  a4.Shift(0,15);
  // Adds 15 to the valid index range along rank 0

  a4.Shift(bottom);
  // Adds the elements of 'bottom' to the valid index range

  a4.Restrict(1, 4, 10);
  // Restricts index range to the intersection of current index range
  // and [5:14] along rank 1

  Ix<2> upper(8,10);
  Ix<2> lower(2,4);
  a4.Restrict(lower,upper);
  // Restricts index range to the intersection of current index range
  // and [2:7,4:9]

  a4.RestrictL(0, 2);
  a4.RestrictL(lower);
  a4.RestrictU(0,8);
  a4.RestrictU(upper);
  // Same as above, except that only the lower (upper) index bound
  // is changed

  a4.Permute(0, 1);
  // Permutes the two ranks 0 and 1
  a4.Permute(Ix<2>(1,0));
  // Permutes the ranks according to the Ix<2> parameter permutation

  a4.Permute(1);
  // Permutes the ranks cyclically 1 step 

  a4.Flip(0);
  // Flips the elements along rank 0. of old index range was [l:u] the
  // new index range is [-u:-l]

  a4.Reverse(0);
  // As Flip(0) but the index range is unchanged

  a4.Restride(0, 2);
  // The array is set to point to every 2'nd element along rank 0

  a4.Restride(Ix<2>(1,3));
  // All ranks are restrided accorgint to the Ix<2> parameter
}
