/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998,2004 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_PERMUTATION_HDR
#define EXPRESSO_PERMUTATION_HDR

#include "debug.h"
#include "tmpltool.h"
#include "ixtype.h"

namespace esso {
  /// Represents a permutation of dimensions
  /** If n is a dimension, then the transformed dimension is
   * p[n] where p is a Permutation object.
   */
  template<int N>
  class Permutation: protected Ix<N,DIMT> {
  public:
    template<int NN>
    friend class Permutation;

    ///@name construction
    //@{
    /** Default constructor. Creates unity permutation. */
    inline Permutation() {
      for(int n=0; n<N; n++)
	this->d[n]= n;
    }
    /** Create a permutation that is a cyclic rotation of dimensions, 
     * that is p[n] == (n+r) modulo N.
     * Negative values for \a r are allowed
     */
    explicit inline Permutation(int r) {
      if(r>=0)
	for(int n=0; n<N; n++)
	  this->d[n] = (n+r)%N;
      else
	for(int n=0; n<N; n++)
	  this->d[n] = N-1-(N-1-(n+r))%N;
    }
    //@}
    ///@name const
    //@{
    /**@return permuted dimension for dimension \a n */
    inline DIMT operator[](DIMT n) const { 
      return this->d[n]; 
    }
    /**@return true iff this object is equal to \a src */
    inline bool operator == (const Permutation<N> &src) const {
      return Ix<N,DIMT>::operator == (src);
    }
    /**@return true iff this object is not equal to \a src */
    inline bool operator != (const Permutation<N> &src) const {
      return Ix<N,DIMT>::operator != (src);
    }
    /** Transform an Ix object by this permutation. I.e.
     * if n is a dimension then
     *  (p*i)[p[n]] = i[n] 
     *
     * Note that this works just like matrix-vector multiplication
     * where the Permutation object acts like a permutation matrix
     * with respect to multiplication.
     *@return vector \a i transformed into permuted space 
     */
    template<class TT>
    inline Ix<N,TT> operator *(const Ix<N,TT> &i) const { 
      Ix<N,TT> tmp;
      for(int n=0; n<N; n++)
	tmp[this->d[n]]=i[n];
      return tmp;
    }
    /** Compose a Permutation from two objects. I.e. if 
     * if p1 and p2 are permutations and n is a dimension then
     * (p1*p2)[n] == p1[p2[n]]
     * Note that this works just like matrix-matrix multiplication
     * where the Permutation object acts like a permutation matrix
     * with respect to multiplication.
     *@return product of two permutations i.e.
     */
    inline Permutation<N> operator *(const Permutation<N> &p) const { 
      Permutation<N> tmp;
      for(int n=0; n<N; n++)
	tmp.d[n]=this->d[p[n]];
      return tmp;
    }
    /** Transform a permutation by another permutation. I.e. if
     * p1 and p2 are permutations and n is a dimension then
     * (p1[p2])[n] == p1[p2[~p1[n]]]
     *
     * Note that this means that p1[p2] is equivalent to p1*p2*~p1.
     *@return permutation \a p transformed into permuted space i.e.
     */
    inline Permutation<N> operator [](const Permutation<N> &p) const { 
      Permutation<N> tmp;
      for(int n=0; n<N; n++)
	tmp.d[this->d[n]]=this->d[p[n]];
      return tmp;
    }
    /** Find out if this is an 'even' permutation, that is if it can
     * be composed by an even number of 'elementary permutations'.
     * An elementary permutation in this context is a permutation of two
     * consecutive dimensions e.g. (1,2)->(2,1).
     *
     * Knowing the 'even' property of a permutation is important in many
     * applications.
     *@return true iff this object is an 'even' permutation.
     */    
    inline bool IsEven() const {
      Permutation tmp(*this);
      bool iseven = true;
      for(int n=0; n<N-1; n++) {
	if(tmp[n]!=n) {
	  int m=n+1;
	  while(tmp[m]!=n)
	    m++;
	  tmp.d[m] = tmp[n];
	  tmp.d[n] = n;
	  iseven = !iseven;
	}
      }
      return iseven;
    }
    //@}
    ///@name nonconst
    //@{
    /** Create object.and checks so that arguments are ok. On failure the
     * contents of this object is unchanged.
     *@return true iff successful
     */
    inline bool Setup(const Ix<N,DIMT> &src) {
      Ix<N,DIMT> tst(0);
      for(int n=0; n<N; n++) {
	if(src[n]<0 ||src[n]>=N)
	  return false;
	if(tst[n]++>0)
	  return false;
      }
      Ix<N,DIMT>::operator =(src);
      return true;
    }
    /** Create object as single permutation of two dimensions.
     * Checks so that arguments are ok. On failure this object is unchanged.
     *@return true iff successful
     */
    inline bool SetupSwap(DIMT n1, DIMT n2) {
      if(n1<0 || n1>=N || n2<0 || n2>=N)
	return false;
      for(int n=0; n<N; n++)
	this->d[n] = n;
      this->d[n1] = n2;
      this->d[n2] = n1;
      return true;
    }
    /** Set this object to it's own inverse
     *@return reference to this object
     */
    inline Permutation<N> &Invert() {
      Permutation<N> tmp;
      for(int n=0; n<N; n++)
	tmp.d[this->d[n]]=n;
      return operator =(tmp);
    }
    /**@return A permutation which is the inverse of this permutation.
     *@see Permutation::Inverse
     */
    inline Permutation<N> operator ~() const {
      return Permutation(*this).Invert();
    }
    /** Concatenate two permutations.
     * if p1 has dim n1 and p2 has dim n2 then
     * (p1|p2)[n] == p1[n] if n<n1 and p2[n-n1]+n1 otherwise
     *@return new permutation object
     */
    template<int N1>
    inline Permutation<N> &Concat(const Permutation<N1> &p1, 
				  const Permutation<N-N1> &p2) {
      for(int n=0; n<N1; n++)
	this->d[n] = p1[n];
      for(int n=0; n<N-N1; n++)
	this->d[n+N1] = p2[n]+N1;
      return *this;
    }
    /**@return A permutation which is the concatenation of this object 
     * with \a p2
     *@see Permutation::Concat
     */
    template<int M>
    inline Permutation<N+M> operator | (const Permutation<M> &p2) const {
      return Permutation<N+M>().Concat(*this,p2);
    }
    //@}
  };
}

#endif
